:- ensure_loaded('DCG-tokenize.pl').

%Ex. ?- check_grammar('XTT_2/therm-rt.pl').
check_grammar(File) :-
	tokenize_file(File,Tokens),
	hmr(Tokens,[]).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%% HMR GRAMMAR %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

hmr --> required_defs(types_def),
	optional_defs(types_grp_def),
	required_defs(attrs_def),
	optional_defs(attrs_grp_def), 
	required_defs(schms_def),
	optional_defs(states_def),
	required_defs(rules_def).

%Here some modifications required. If base is numeric then can be numbering without ordered. 
%In any other case numbering requires domain to be ordered.
	
types_def --> [xtype],['['],
	type_name,[,],
	type_base,[,],
	optional_field(type_scale),
	optional_field(type_order),
	type_domain,
	optional_field(type_description),
	[']'],[.],!.
	

type_name --> [name],[:],str.
type_base --> [base],[:],[symbolic].
type_base --> [base],[:],[numeric].
type_domain --> [domain],[:],lst(domain).
type_scale --> [scale],[:],[ss],[,].
type_order --> [ordered],[:],[yes],[,].
type_order --> [ordered],[:],[no],[,].
type_description --> [,],[desc],[:],str.


attrs_def --> [xattr],['['],
	attr_name,[,],
	attr_class,[,],
	attr_type,[,],
	attr_comm,
	optional_field(attr_callback),
	optional_field(attr_abbrev),
	optional_field(attr_description),
	[']'],[.],!.
	
attr_name --> [name],[:],str.
attr_class --> [class],[:],[simple].
attr_class --> [class],[:],[general].
attr_type --> [type],[:],str.
attr_comm --> [comm],[:],[in].
attr_comm --> [comm],[:],[out].
attr_comm --> [comm],[:],[inter].
attr_comm --> [comm],[:],[comm].
attr_callback --> [,],[callback],[:],str.
attr_abbrev --> [,],[abbrev],[:],str.
attr_description --> [,],[desc],[:],str.


schms_def --> [xschm],
	schm_name,
	optional_field(schm_description), 
	[:], 
	schm_from,
	[=],[=],[>],
	schm_to,[.],!.
	
schm_name --> str.
schm_description --> [/],str.
schm_from --> lst(general).		
schm_to --> lst(general).			

	
rules_def --> [xrule],
	rule_id,[:],
	['['],rule_conditions,[']'],
	[=],[=],[>],
	['['],rule_decisions,[']'],
	optional_field(rule_actions),
	optional_field(rule_link),[.],!.
	
rule_actions --> [*],[*],[>],
	['['],rule_operations,[']'],!.

rule_id --> str,[/],str.

rule_conditions --> rule_condition,[,],rule_conditions.
rule_conditions --> rule_condition.
rule_condition --> str,operator,parameter.
operator --> [eq].
operator --> [neq].
operator --> [in].
operator --> [lt].
operator --> [gt].
operator --> [notin].
operator --> [subset].
operator --> [supset].
operator --> [sim].
operator --> [notsim].
parameter --> str.
parameter --> lst(general).



rule_decisions --> rule_decision.
rule_decisions --> rule_decision,[,],rule_decisions.
rule_decision --> str,[set],str.

rule_operations --> rule_operation,rule_operations.
rule_operations --> rule_operation.

rule_operation --> str,['('],lst_arg(general),[')'].

rule_link --> [:],str.
rule_link --> [:],str,[/],str.

types_grp_def --> [xtpgr],['['],
	[type_grp_name],[:],str,[,],
	optional_field(type_grp_abbrev),
	[types],[:],lst(general),
	[']'],[.].
	
type_grp_abbrev --> [abbrev],[:],str,[,].

attrs_grp_def --> [xatgr],['['],
	[attr_grp_name],[:],str,[,],
	optional_field(attr_grp_abbrev),
	[attrs],[:],lst(general),
	[']'],[.].
	
attr_grp_abbrev --> [abbrev],[:],str,[,].
	
states_def --> [xstat],str,[:],['['],
	str,[,],parameter,[']'],[.].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%% ADDITIONAL DEFS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

str --> [String],{atom(String)},!.
lst(Style) --> ['['],lst_args(Style),[']'],!.
lst_args(Style) --> lst_arg(Style).
lst_args(Style) --> lst_arg(Style), [,],lst_args(Style).
lst_arg(domain) -->  lst_arg(general),[/],[Weight],{number(Weight)}.
lst_arg(domain) -->  lst_arg(general).
lst_arg(general) --> range.
lst_arg(general) --> [Atom], {atom(Atom)}.
lst_arg(general) --> lst(general).

range --> [LowerLimit],[to],[UpperLimit],
	{atom_number(LowerLimit,_),atom_number(UpperLimit,_)}.

required_defs(Def) --> Def, required_defs(Def),!.
required_defs(Def) --> Def.

optional_defs(Def) --> Def,!,optional_defs(Def).
optional_defs(_) --> [].

optional_field(Field) --> Field.
optional_field(_)--> [].




