%
% $Id: heart-alsv.pl,v 1.11 2009-05-13 10:23:50 esimon Exp $
%
% Copyright 2004-9 by Grzegorz J. Nalepa and A. Ligeza
%
% HeaRT the ALSV(FD) handling
%
%
%     Copyright (C) 2004-6 by G. J. Nalepa, A. Ligeza
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     HeaRT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of HeaRT.
%
%     HeaRT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     HeaRT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with HeaRT.  If not, see <http://www.gnu.org/licenses/>.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- write('HeaRT: $Id: heart-alsv.pl,v 1.11 2009-05-13 10:23:50 esimon Exp $'),nl.

%%% Finite domain declaration
%%% domain(<name>,<list-of-values>).

%%% Set value definition
%%% set(<name>,<list-of-values>).

%%% Ordered Set value definition
%%% set(<name>,<list-of-values>,<order-predicate-name>).

%%% Attribute declaration
%%% attribute(<attribute-name>,<attribute-type>,<attribute-domain>).
% attribute(aDD,atomic,day_of_week).
% attribute(sDD,set,day_of_week).


%%% State definition: by listing all the values of the attributes
%%% state(<state-identifier>,<attribute>,<value>,<type>).
%%% state_time(<state-identifier>,<time>).
%%% <type> := atomic|set|...\
% state(s1,aMO,atomic,march).
% state(s2,sTM,set,[7,8,9,10,11,12,13,14,15]).


%%% Definition of facts: to be used in rule preconditions
%%% f(<attribute-type>,<attribute-name>,<relation>,<attribute-domain>)
%%% <attribute-type> := atomic|set|....
%%% For <attribute-type>=atomic <relation> := \eq, \neq, \in, \not\in
%%% For <attribute-type>=set <relation> := \eq, \neq, \subseteq,\supseteq,\sim, \not\sim

%%% To be defined: 'NULL','_',[]


%%% Satisfaction of a single fact versus state: a fact 
%%% f(<attribute-type>,<attribute-name>,<relation>,<attribute-domain>)
%%% is satisfied within some state <state> iff <relation> is valid within this state;
%%% Deatils are defined by the first rows of table1 and table2.
%%% valid(<fact>,<state>).

%%% Table1: cases of \eq, \neq, \in, \notin versus atomic StateValue in State
%valid(f(atomic,A,eq,Value),State) :-
%	state(State,A,atomic,StateValue),
%	Value == StateValue,
%	!.  
%valid(f(atomic,A,neq,Value),State) :-
%	state(State,A,atomic,StateValue),
%	Value =\= StateValue,
%	!. 
%valid(f(atomic,A,in,SetValue),State) :-
%	state(State,A,atomic,StateValue),
%	member(StateValue,SetValue),
%	!. 
%valid(f(atomic,A,notin,SetValue),State) :-
%	state(State,A,atomic,StateValue),
%	\+member(StateValue,SetValue),
%	!. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% heart

alsv_valid(Att in [L to U], State) :-
	alsv_attr_class(Att,simple),
	xstat State: [Att,StateValue],
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[L]), alsv_values_check(Att,[U]),
	StateValue >= L,
	StateValue =< U,	
	heart_debug(2,['Valid:',Att,in,[L to U]]).
	
alsv_valid(Att notin [L to U], State) :-
	alsv_attr_class(Att,simple),
	xstat State: [Att,StateValue],
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[L]), alsv_values_check(Att,[U]),
	StateValue >= U,
	StateValue =< L,	
	heart_debug(2,['Valid:',Att,notin,[L to U]]).
	
alsv_valid(Att in SetValue, State) :-
	alsv_attr_class(Att,simple),
	xstat State: [Att,StateValue],
	!,
	alsv_values_check(Att,[StateValue]), 	
	alsv_values_check(Att,SetValue),
	%normalize(Att,[StateValue],[NormStateValue]),
	%normalize(Att,SetValue,NormSetValue),
	%member(NormStateValue,NormSetValue),
	member(StateValue,SetValue),
	heart_debug(2,['Valid:',Att,in,SetValue]).
	
alsv_valid(Att notin SetValue, State) :-
	alsv_attr_class(Att,simple),
	xstat State: [Att,StateValue],
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,SetValue),
	\+ member(StateValue,SetValue),
	heart_debug(2,['Valid:',Att,notin,SetValue]).
	
alsv_valid(Att eq Value, State) :-
	alsv_attr_class(Att,simple),
	xstat State: [Att,StateValue],
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	Value == StateValue,
	heart_debug(2,['Valid:',Att,eq,Value]).
	
alsv_valid(Att eq any, State) :-
	alsv_attr_class(Att,simple),
	xstat State: [Att,_],
	!,
	alsv_values_check(Att,[StateValue]), %alsv_values_check(Att,[SetValue]),
	Value == StateValue,
	heart_debug(2,['Valid:',Att,eq,any]).
	
alsv_valid(Att neq Value, State) :-
	alsv_attr_class(Att,simple),
	xstat State: [Att,StateValue],
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	Value =\= StateValue,
	heart_debug(2,['Valid:',Att,neq,Value]).

alsv_valid(Att gt Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),	
	xstat State: [Att,StateValue],
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	Value < StateValue,heart_debug(2,['watafak',Value,StateValue]),
	heart_debug(2,['Valid:',Att,gt,Value]).
alsv_valid(Att lt Value, State) :-
	alsv_attr_class(Att,simple),
	alsv_domain(Att,_,numeric),
	xstat State: [Att,StateValue],
	!,
	alsv_values_check(Att,[StateValue]), alsv_values_check(Att,[Value]),
	Value > StateValue,
	heart_debug(2,['Valid:',Att,gt,Value]).
	
%%% Table2: cases of \eq, \neq, \subseteq, \supseteq,\sim, \notsim  versus set StateValue in State
alsv_valid(Att eq Set,State) :-
	alsv_attr_class(Att,general),
	xstat State: [Att,StateValue],
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	eqset(Set,StateValue),
	heart_debug(2,['Valid:',Att,eq,Set]).
	
alsv_valid(Att neq Set,State) :-
	alsv_attr_class(Att,general),
	xstat State: [Att,StateValue],
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	neqset(Set,StateValue),
	heart_debug(2,['Valid:',Att,eq,Set]).
	
alsv_valid(Att subseteq Set,State) :-
	alsv_attr_class(Att,general),
	xstat State: [Att,StateValue],
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	subset(StateValue,Set),
	heart_debug(2,['Valid:',Att,subset,Set]).
	
alsv_valid(Att supseteq Set,State) :-
	alsv_attr_class(Att,general),
	xstat State: [Att,StateValue],
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	subset(Set,StateValue),
	heart_debug(2,['Valid:',Att,supset,Set]).
	
alsv_valid(Att sim Set,State) :-
	alsv_attr_class(Att,general),
	xstat State: [Att,StateValue],
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	intersection(StateValue,Set,[_|_]),
	heart_debug(2,['Valid:',Att,sim,Set]).
	
alsv_valid(Att notsim Set,State) :-
	alsv_attr_class(Att,general),
	xstat State: [Att,StateValue],
	!,
	alsv_values_check(Att,StateValue),
	alsv_values_check(Att,Set),
	intersection(StateValue,Set,[]),
	heart_debug(2,['Valid:',Att,notsim,Set]).


%valid(f(set,A,\subseteq,SetValue),State) :-
%	state(State,A,set,StateValue),
%	subset(SetValue,StateValue),
%	!.
%valid(f(set,A,\supseteq,SetValue),State) :-
%	state(State,A,set,StateValue),
%	subset(StateValue,SetValue),
%	!.
%valid(f(set,A,\sim,SetValue),State) :-
%	state(State,A,set,StateValue),
%	intersect(SetValue,StateValue,[_|_]),
%	!.
%valid(f(set,A,\notsim,SetValue),State):-
%	state(State,A,set,StateValue),
%	intersect(SetValue,StateValue,[]),
%	!.
	

alsv_valid(F,_) :-
	heart_debug(2,['Check for:',F,'NYI!']),
	fail.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% alsvd checks
alsv_values_check(_,[]).

%set state value check
alsv_values_check(Att,[LS to US|Rest]) :-
	alsv_domain(Att,[LD to UD],numeric),
	heart_debug(2,['Check:',Att,numericreg,LS to US]),		
	LS >= LD,
	US =< UD,	
	alsv_values_check(Att,Rest),
	!.
	
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),
	alsv_domain(Att,[LD to UD],numeric),
	heart_debug(2,['Check:',Att,numericreg,StateValue]),
	number(StateValue),
	StateValue >= LD,
	StateValue =< UD,	
	alsv_values_check(Att,Rest),
	!.
	
	
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),	
	alsv_domain(Att,Domain,numeric),
	heart_debug(2,['Check:',Att,numeric,StateValue]),	
	member(StateValue,Domain),
	alsv_values_check(Att,Rest),
	!.
	
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	alsv_order(Att,yes),
	member(X/StateValue,Domain),
	alsv_values_check(Att,Rest),
	!.
	
alsv_values_check(Att,[StateValue|Rest]) :-
	nonvar(StateValue),		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	alsv_order(Att,yes),
	member(StateValue/X,Domain),
	alsv_values_check(Att,Rest),
	!.
	
alsv_values_check(Att,[StateValue|Rest]) :-
%	nonvar(StateValue),		
	alsv_domain(Att,Domain,symbolic),
	heart_debug(2,['Check:',Att,symbolic,StateValue]),
	member(StateValue,Domain),
	alsv_values_check(Att,Rest),
	!.

alsv_values_check(Att,[V|Rest]) :-
	write('  ERROR: values of \"'), write(Att),
	write('\": '), write(V), write(' NOT in the domain!'),nl.

%retrieves attribute domain, and base: symbolic, numeric
alsv_domain(Att,Domain,Base) :-
	xattr(A),
	member(name:Att,A),
	member(type:T,A),
	xtype(Ts),
	member(name:T,Ts),
	member(domain:Domain,Ts),
	member(base:Base,Ts).

% retrieves attribute class: simple, general
alsv_attr_class(Att,Class) :-
	xattr(A),
	member(name:Att,A),
	member(class:Class,A).

alsv_order(Att,Order) :-
	xattr(A),
	member(name:Att,A),
	member(type:T,A),
	xtype(Ts),
	member(name:T,Ts),
	member(ordered:Order,Ts),!.

alsv_order(Att,no).
	
normalize(Att, Value, Output) :-
	alsv_order(Att,no),
	Value = Output.

normalize(Att, Value, Output) :-
	alsv_order(Att,yes),
	findall(Numbering,member(Name/Numbering,Value),Output).


%%% Auxiliary predictaes
%%% Set equality definition:
eqset(Set1,Set2) :-
	subset(Set1,Set2),
	subset(Set2,Set1).
%%% Set in-equality definition:
neqset(Set1,Set2) :-
	\+ eqset(Set1,Set2).
	
% heart
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



% heart
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

alsv_transition(Att set Val) :-
%	atom(Val),
	\+ Val = [_|_],
	alsv_attr_class(Att,simple),
%	alsv_values_check(Att,[Val]),
	alsv_new_val(Att,Val).

alsv_transition(Att set Val) :-
%	atom(Val),
	Val = [_|_],
	alsv_attr_class(Att,general),
%	alsv_values_check(Att,[Val]),
	alsv_new_val(Att,Val).

% fixme: no direct assert/retract, but set/get_att_val
alsv_new_val(A,V):-
%	xstat current:[A,O],
	retractall(xstat current:[A,_]),
	assert(xstat current:[A,V]),
	heart_debug(2,['Trans:',A,to,V]).
%	heart_debug(['Transition: ',A,from,O,to,V]).

%alsv_transition(Att set Val) :-
%	Val = [_|_],
	
%  [therm_set set 24].
%  [therm_set set [24,5]].

%  [therm_set set seval union([24,5],intersect('att_ala',[4 to 8])). EXCEPT COMPLEMENT? % binary relations! result=set

%  [therm_set set aeval 4 + 7 * 'att_ala']. %dowolone wyr oblicznae przez is, maths builtin, defined funcs? % result=number binary relations! therm_set in numeric

  
%%% Checking if a list of facts -- precondition formula is satisfied 
%%% satisfied(<list-of-facts>,State).

%satisfied([],_) :- !.
%satisfied([Fact|List],State) :-
%	valid(Fact,State),
%	satisfied(List,State).

%%% Auxiliary predictaes
%%% Set equality definition:
%eqset(Set1,Set2) :-
%	subset(Set1,Set2),
%	subset(Set2,Set1).
%%% Set in-equality definition:
%neqset(Set1,Set2) :-
%	\+ eqset(Set1,Set2).




