%
% $Id: heart-infer.pl,v 1.1 2009-05-07 22:18:53 gjn Exp $
%
%     Copyright 2004-9 by Grzegorz J. Nalepa and A. Ligeza
%
%     HeaRT inference procedures
%
%
%     Copyright (C) 2004-6 by G. J. Nalepa, A. Ligeza
%     Copyright (C) 2009   by S. Bobek, M. Gawedzki
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     HeaRT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of HeaRT.
%
%     HeaRT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     HeaRT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with HeaRT.  If not, see <http://www.gnu.org/licenses/>.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- write('HeaRT: $Id: heart-infer.pl,v 1.1 2009-05-07 22:18:53 gjn Exp $'),nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Single Table processing
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% table processing
runtables([]).
runtables([R|T]) :-
	runtable(R),
	runtables(T).

runtable(T) :-
%	write('*Firing tabl: \"'),
%	write(T), write('\" ('),
	%FIXME: conditionless and decisionless rules support!!!
	xschm(T:X==>Y),
%	write(X),write('==>'),write(Y),write(')'),nl,
	heart_debug(1,['Firing tabl: ','(',T,')',X,'==>',Y]),
	runrules(T),
	nl.

runrules(T):-
	xrule(T/R:C==>D),
%	write(' *Firing rule: '), write(T/R),nl,
	heart_debug(1,['Firing rule: ',T/R]),
	cond_satisfy(C),
	process_decision(D,Go),
	inference_transfer(Go),
	fail.
runrules(_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% GDI-Firing
%inference: GoalDriven GDI
%
%    * get set of goal tables
%    * build a sequential execution plan, by backward chaining using TT links, list of needed components
%    * run independent tables
%    * wave-like in steps
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Goal driven for specified tables. Tables shoul be a list/set ex.: runrules_GDI([os]).
runrules_GDI(Goals) :-
	create_stack(OrderedTables,Goals),
	runtables(OrderedTables).
	
% Runs all tables in order they are linked
runrules_GDI :-
	get_goal_tables(Goals),
	write('Goals: '), write(Goals),nl,
	create_stack(OrderedTables, Goals),
	runtables(OrderedTables).

create_stack(Stack,Goals) :-
	get_pred_tables(Goals,Predecessors),
	join_tables(Predecessors,Goals,Stack).

get_pred_tables([],[]).
get_pred_tables(Goals,Predecessors) :-
	findall(C,condition_goal(C,Goals),Conds),
	flatten(Conds,FC),
	findall(T,decision_goal(T,FC),Preds),
	flatten(Preds,FP),
	get_pred_tables(FP,FPP),
	join_tables(FPP,FP,Predecessors), !.

join_tables([],X,X).
join_tables([X|L1],L2,[X|L3]) :-
	join_tables(L1,L2,L3).


condition_goal(C,Set) :-
	xschm(A:C==>_),
	member(A,Set).
		
decision_goal(T,Set) :-
	xschm(T:_==>D),
	ord_intersect(D,Set). %If we assume that there may be links to more than one table.

find_goal_atts(GoalAtts) :-
	findall(D,xschm(_:_==>D),Decs),
	findall(C,xschm(_:C==>_),Conds),
	flatten(Decs,FD),
	flatten(Conds,FC),
	subtract(FD,FC,GoalAtts).
	
get_goal_tables(GTables) :-
	find_goal_atts(GA),
	findall(I,list_goaltabs(GA,I),GTables).
	
list_goaltabs(GA,I) :-	
	xschm(I:_==>DN),
	subset(DN,GA).
	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Token-Transfer-Firing

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Create stack, but fill also create database containg data about inputs, (num of tokens tables require)
% Number of tokens are determined by number of R->T liks (not T->T as previously)

runrules_token_transfer :-
	create_stack(Stack),
	fill_token_knowledge(Stack),
	run_token_inference(Stack).
	
fill_token_knowledge([]).
fill_token_knowledge([Head|Tail]) :-
	get_reqtok_number(Head,Number),
	assert(xtable(Head,0,Number)),
	fill_token_knowledge(Tail).
	
% +Table -Number
% get_reqtok_number(Table,Number) :-
%	findall
