%
% $Id: heart.pl,v 1.7 2009-05-13 14:21:44 gjn Exp $
%
%     Copyright 2004-9 by Grzegorz J. Nalepa and A. Ligeza
%
%     HeaRT the main shell
%
%
%     Copyright (C) 2004-6 by G. J. Nalepa, A. Ligeza
%     Copyright (C) 2009   by S. Bobek, M. Gawedzki
%     Copyright (C) 2006-9 by the HeKatE Project
%
%     HeaRT has been develped by the HeKatE Project, 
%     see http://hekate.ia.agh.edu.pl
%
%     This file is part of HeaRT.
%
%     HeaRT is free software: you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation, either version 3 of the License, or
%     (at your option) any later version.
%
%     HeaRT is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
%
%     You should have received a copy of the GNU General Public License
%     along with HeaRT.  If not, see <http://www.gnu.org/licenses/>.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- write('HeaRT: $Id: heart.pl,v 1.7 2009-05-13 14:21:44 gjn Exp $'),nl.
:- write('\nCopyright (C) by G. J. Nalepa and A. Ligeza 2004-2006 and the HeKatE Project 2006-2009\nHeaRT is free software: you can redistribute it and/or modify it under\nthe terms of the GNU General Public License as published by the Free Software Foundation,\neither version 3 of the License, or (at your option) any later version.\n\n').

:-dynamic(xstat/1).

flag_debug(3).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TODO!!
% 20.02.2009
% - alsv(fd) support as of the amcs paper
% - intertable inference algorithms as of the amcs paper
% - atts blackboard
% - attribute domains, etc. specs!

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Operators
%
% structural operators
:-op(830,fx,xrule).         % used to define rule
:-op(830,fx,xtype).         % used to define type
:-op(830,fx,xattr).         % used to define attribute
:-op(830,fx,xschm).         % used to define scheme
:-op(830,fx,xstat).         % used to define state
:-op(830,fx,xtpgr).         % used to define group of attributes
:-op(830,fx,xatgr).         % used to define group of attributes
%
:-op(800,xfx,==>).		% used to separate LHS and RHS of rule
:-op(500,xfx,**>).		% used to indicate the actions in the decision part
%:-op(800,xfx,if).		% used to separate LHS and RHS of rule
%:-op(800,xfx,then).		% used to separate LHS and RHS of rule
%:-op(800,xfx,go).		% used to separate LHS and RHS of rule

%%:-op(700,xfy,#).            % used for unification instead of =
%%:-op(500,xfy,:).            % used to separate attributes and values, this is defined in ISO!
%% 400	yfx	*, /,

% alvs operators
% table 1
:-op(300,xfy,eq).
:-op(300,xfy,neq).
:-op(300,xfy,in).
:-op(300,xfy,notin).
% table2
%:-op(300,xfy,eq).
%:-op(300,xfy,neq).
:-op(300,xfy,subseteq).
:-op(300,xfy,supseteq).
:-op(300,xfy,sim).
:-op(300,xfy,notsim).

:-op(816,xfx,to).  % used to specify ranges for general ordered atts, as <mon,fri> [mon to fri]
:-op(300,xfy,set). % setting val in decision

:-op(301,xfy,gt).
:-op(301,xfy,lt).

%:- dynamic(xrule).
%:- dynamic(xtype).
%:- dynamic(xattr).

:-dynamic(xtable/3). % facts about tokens for token inference

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Includes
:- ensure_loaded('heart-alsv.pl').
:- ensure_loaded('heart-infer.pl').
:- ensure_loaded('heart-rules.pl').
:- ensure_loaded('heart-state.pl').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Help
%
hlp :-
	nl,
	write('load an example, e.g. [therm-rt].'), nl,
	write('gox. - semiautomatic inference based on table order.'),nl,
	write('gox(State,Tables). - start from given state, inference based on explicit table list.'),nl,	
	write('hlt. - goodbye.'), nl,
	write('FIXME'), nl,
	nl.
:-
	nl,
	write('This is HeaRT, the HeKatE Run Time engine.'), nl,
	write('Use \'hlp.\' to get some help'), nl,
	nl.

hlt :- halt.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Main shell predicates

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% gox in det.
%
%  Runs the Heart XTT interpreter
%
gox :-
	% check/bootstrap attributes
	% built a set of runable tables
	init_tables(RT),
	runtables(RT).
	% maybe start from initial state?
	% get a
%	.

gox(State,T) :-
	% check/bootstrap attributes
	% built a set of runable tables
	state_print(State),
	state_copy(State,current),
	runtables(T),
	state_print(current).


goxio(State,T) :-
	% check/bootstrap attributes
	% built a set of runable tables
	findall_input_atts(IA), %all atts having 'comm: in'
	fireall_callbacks(IA),  %fire all callbacks in given atts
	state_print(State),
	state_copy(State,current),
	runtables(T),
	findall_output_atts(OA),%all atts having 'comm: out'
	fireall_callbacks(OA),	
	state_print(current).

table_print :-
	write('Tables: '),
	table_print_helper.
table_print_helper :-		
	xschm T:_==>_,
	write(T),write(', '),
	fail.
table_print_helper.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% knowledge base syntax and structure checks
%
xtype_def([
	   [name, [_]],
	   [base, [symbolic,numeric]],
	   [ordered,[yes,no]],
	   [domain,[_]],
	   %[abbrev,[_]],
	   [desc, [_]],
	   [numbering,[_]],
	   [length, [_]],
	   [scale,[_]]	  
	  ]
	  ).

xtype_check :-
	xtype(T),
	member(name:N,T),
	write('Checking xtype: "'), write(N), write('"  '), nl,
	xtype_check_single(T),
	nl,
	fail.
xtype_check :-
	write('No more xtypes to check.').

xtype_check_single(T) :-
	member(D:V,T),
	xtype_def(Y),
	member([D,_],Y),
	write('    '),write(D:V), write(' -> OK').

xattr_def([
	   [name,[_]],
	   [class,[simple,general]],
	   [type,[_]],
	   [comm,[in,inter,out,comm]],
	   [abbrev,[_]],
	   [desc,[_]],
	   [callback,[_]]
	  ]).

%xrules_check :-.
% scheme exists
% attributes valid with scheme
% attributes defined
% on the runtimelevel: see if val is in the att domain!
% if control spec exists, check table and rule id
